#* Functions for handling the Network Data Dictionary (NDD) of PowerModels.jl
#*------------------------------------------------------------------------------

#=
Reads the geographic bus locations (longitude and latitude coordinates in degrees) from a CSV file and adds them to the NDD. Returns the resulting NDD, in which each bus has a "bus_lat" and "bus_lon" entry. The columns containg the longitude and latitude coordinates in the CSV file should be called "SubLongitude" and "SubLatitude", respectively.
=#
function addlocs!(
        network_data::Dict{String,<:Any}, 
        csvfile::String
    )

    pos = Dict{String,Any}(
        "baseMVA"  => 100.0,
        "bus" => Dict{String,Any}(), 
        "per_unit" => true
    )

    BusData::DataFrame = CSV.File(csvfile) |> DataFrame!
    N = size(BusData, 1) # number of buses
    sizehint!(pos["bus"], N)

    for bus in eachrow(BusData)
        pos["bus"][string(bus[:Number])] = Dict(
            "bus_lon" => bus[:SubLongitude], 
            "bus_lat" => bus[:SubLatitude]
        )
    end
    update_data!(network_data, pos) # add positions to network_data
    
    return network_data
end

#=
Returns a dictionary containing geographic bus locations that are included in the NDD. The dicitonary structure is bus-index => (lon, lat). If getlims=true, the minmimum and maximum longitude and latitude coordinates are returned as well. Can be useful for plotting the grid.
=#
function getlocs(network_data::Dict{String,<:Any}; getlims=false)
    pos = Dict(
        parse(Int64, i) => (b["bus_lon"], b["bus_lat"]) 
        for (i, b) in network_data["bus"]
    )

    ### Check whether to return lon- and lat-limits or not
    if getlims == true
        N = length(network_data["bus"]) # number of buses
        lats = [
            network_data["bus"]["$i"]["bus_lat"] 
            for i in keys(network_data["bus"])
        ] # latitudes
        lons = [
            network_data["bus"]["$i"]["bus_lon"] 
            for i in keys(network_data["bus"])
        ] # longitudes
        xlims = (minimum(lons), maximum(lons))
        ylims = (minimum(lats), maximum(lats))
        return pos, xlims, ylims
    else
        return pos
    end
end

#*------------------------------------------------------------------------------

#=
Returns the bus types of all buses in the NDD in form a dictionary. The different bus types are "load", "gen" (generator), "load_and_gen" and "empty".
=#
function get_bustypes(network_data::Dict{String,<:Any})
    ### Active loads
    loads = unique(
        [load["load_bus"] for (i, load) in network_data["load"] 
        if load["status"] == 1]
    )
    
    ### Active generators
    gens = unique(
        [gen["gen_bus"] for (i, gen) in network_data["gen"] 
        if gen["gen_status"] == 1]
    )
    
    ### Buses with both load and generation
    load_and_gen = [i for i in loads if i in gens]
    
    ### Remove buses with load and generation from pure loads and generators
    filter!(i -> i ∉ load_and_gen, loads)
    filter!(i -> i ∉ load_and_gen, gens)
    
    ### Empty buses with neither load nor generation
    empty = unique(
        [bus["index"] for (i, bus) in network_data["bus"] 
        if bus["index"] ∉ vcat(loads, gens, load_and_gen)]
    )

    ### Check whether the right number of bus types was obtained
    sum = (length(loads) + length(gens) + length(load_and_gen) + length(empty))
    N::Int64 = length(network_data["bus"]) # number of buses
    @assert sum == N "$sum bus types were obtained instead of $N"

    return Dict(
        "loads" => loads,
        "gens" => gens,
        "load_and_gen" => load_and_gen,
        "empty" => empty       
    )
end